// Copyright 1994 by Jon Dart.  All Rights Reserved.

#include "moveord.h"
#include "scoring.h"
#include "rmove.h"
#include "constant.h"
#include "attacke.h"
#include "bearing.h"
#include "util.h"
#include "array.h"

static Move Killers[Constants::MaxPly];

inline void swap( Move moves[], const int i, const int j)
{
    Move tmp = moves[j];
    moves[j] = moves[i]; 
    moves[i] = tmp;
}

#define SORT_LIMIT 8

void 
Move_Ordering::order_moves(Board & board,
			   Move moves[], const int num_moves,
			   const int ply, const Move & best_move)
{
    if (num_moves == 0)
        return;
    if (ply > 0)
    {
        // We can't afford a full sort of the moves at deeper plies.
	// So we move promising moves towards the front, then sort
	// the first few moves.
        Array <int> scores(SORT_LIMIT);
	int j = 0;
	int best_index = -1;
	int killer = -1;
	Move tmp;
	for (int i = 0; i < num_moves; i++)
	{
	    if (i < SORT_LIMIT)
	    	scores[i] = 0;
	    if (moves[i] == best_move)
  	        best_index = i;
	    else if (moves[i] == Killers[ply])
		killer = i;
	 }
	 if (best_index >= 0)
	 {
	    scores[0] = 1000;
	    if (best_index > 0)
	       swap(moves,best_index,0);
	    ++j;
	 }
	 if (killer >= 0)
	 {
	    if (killer != j)
               swap(moves,killer,j);
	    scores[j] = 30;
	    ++j;
	 }

	 int start = j;
 	 for (i = start; i < num_moves && j < SORT_LIMIT; i++)
	 {
	    const Square dest = moves[i].DestSquare();
	    Piece capture = board[dest];
	    if (!capture.IsEmpty())
	    {
#ifdef RANGE_CHECK
  	        assert(j < num_moves);
#endif
		if (i != j)
		{
		    swap(moves,i,j);
	        }
		if (j < SORT_LIMIT)
		{
		    Square start = moves[j].StartSquare();
		    int start_val;
		    if (board[start].Type() == Piece::King)
		       start_val = Piece::Value(Piece::Pawn);
		    else
		       start_val = board[start].Value();
		    int capture_val = capture.Value();
   	            if (Scoring::check_en_prise(board,dest,board[dest]))
		       // piece appears to be en prise
		       scores[j] = capture_val;
		    else if (capture_val > start_val)
		       scores[j] = capture_val - start_val;
		    else
		       scores[j] += 30 + (capture_val - start_val)/16;
		}
		else
		    break;
		++j;
	    } 
	}
	if (j > 1)
	    sort_moves(moves,(int*)scores.get_data(),Util::Min(j,SORT_LIMIT));
    } 
    else 
    {
	// ply == 0

	Array<int> scores(num_moves);
	scores[0] = 1000;	// p.v. move

	for (int i = 1; i < num_moves; i++)
	{
	    scores[i] = 0;
	    Square dest_square = moves[i].DestSquare();
	    Square start_square = moves[i].StartSquare();
	    Piece capture = board[dest_square];
	    if (!capture.IsEmpty())
	    {
	        ExtendedMove emove(board,moves[i]);
   	        if (board.num_attacks(dest_square, board.OppositeSide()) == 0)
		    // attack on undefended piece
		    scores[i] += capture.Value();
		else if (emove.Special() == ExtendedMove::Normal)
		{
	            int est = attack_estimate(board,emove);
		    if (est > 0)
		       scores[i] += est;
		    else // losing capture
		       scores[i] += 40 + (est/16);
		}
		else // en-passant?
		    scores[i] += 32;
	    }
	    if (moves[i] == Killers[ply])
		scores[i] += 30;
	    ReversibleMove rmove(board, start_square, dest_square);
	    board.MakeMove(rmove);
	    scores[i] -= Scoring::positional_score(board);
	    if (board.CheckStatus() == Board::InCheck)
	       scores[i] += 25;
	    board.UndoMove(rmove);
	}
	sort_moves(moves, (int*)scores.get_data(), num_moves);
    }
}

void 
Move_Ordering::clear_killer()
{
    for (int i = 0; i < Constants::MaxPly; i++)
	Killers[i].MakeNull();
}

void 
Move_Ordering::set_killer(ExtendedMove & move, const unsigned ply)
{
    Killers[ply] = move;
}

void
Move_Ordering::get_killer(Move & move, const unsigned ply)
{
    move = Killers[ply];
}
